@@ -4,7 +4,7 @@
 
    Copyright (C) 1996-2021 Markus Franz Xaver Johannes Oberhumer
    Copyright (C) 1996-2021 Laszlo Molnar
-   Copyright (C) 2000-2024 John F. Reiser
+   Copyright (C) John F. Reiser
    All Rights Reserved.
 
    UPX and the UCL library are free software; you can redistribute them
@@ -36,12 +36,13 @@
 extern unsigned Pprotect(void *, size_t, unsigned);
 extern void *Pmap(void *, size_t, unsigned, unsigned, int, size_t);
 extern int Punmap(void *, size_t);
+extern int Psync(void const *, size_t, unsigned);
 extern size_t Pwrite(unsigned, void const *, size_t);
+#define MS_SYNC 4
+#define MFD_EXEC 0x10
 
 extern void f_int3(int arg);
 
-#define DEBUG 0
-
 #ifndef DEBUG  //{
 #define DEBUG 0
 #endif  //}
@@ -81,6 +82,15 @@ extern void f_int3(int arg);
 /*und*/ : "x30"); \
     dprintf(r_fmt, args); \
 })
+#elif defined(__riscv) //}{
+#define DPRINTF(fmt, args...) ({ \
+    char const *r_fmt; \
+    asm("jal %0, 0f; .string \"" fmt "\"; .balign 4; 0:" \
+/*out*/ : "=r"(r_fmt) \
+/* in*/ : \
+/*und*/ : ); \
+    dprintf(r_fmt, args); \
+})
 
 #endif  //}
 
@@ -145,7 +155,7 @@ xread(Extent *x, char *buf, size_t count)
     char *p=x->buf, *q=buf;
     size_t j;
     if (x->size < count) {
-        exit(127);
+        err_exit(8);
     }
     for (j = count; 0!=j--; ++p, ++q) {
         *q = *p;
@@ -213,6 +223,7 @@ ERR_LAB
         }
         else { // copy literal block
             DPRINTF("  copy %%p  %%p  %%p\\n", xi->buf, xo->buf, h.sz_cpr);
+            xi->size += sizeof(h);  // xread(xi, &h, sizeof(h)) was a peek
             xread(xi, xo->buf, h.sz_cpr);
         }
         xo->buf  += h.sz_unc;
@@ -236,39 +247,66 @@ ERR_LAB
 /*und*/ : "x30"); \
     str; \
 })
+#elif defined(__riscv) //}{
+#define addr_string(string) ({ \
+    char const *str; \
+    asm("jal %0,0f; .string \"" string "\"; .balign 4; 0:" \
+/*out*/ : "=r"(str) \
+/* in*/ : \
+/*und*/ : ); \
+    str; \
+})
 #else  //}{
        error;
 #endif  //}
 
 extern int memfd_create(const char *name, unsigned int flags);
 
+#define ElfW(sym) Elf64_ ## sym
+
+#define nullptr (void *)0
+
+extern char *upx_mmap_and_fd(  // x86_64 Android emulator of i386 is not faithful
+     void *ptr  // desired address
+     , unsigned len  // also pre-allocate space in file
+     , char *pathname  // 0 ==> call get_upxfn_path, which stores if 1st time
+);
+
 #if defined(__x86_64__)  //{
-static char *
-make_hatch_x86_64(
-    Elf64_Phdr const *const phdr,
+char *
+make_hatch(
+    ElfW(Phdr) const *const phdr,
     char *next_unc,
     unsigned frag_mask
 )
 {
-    char *hatch = 0;
+    unsigned *hatch = 0;
+    unsigned code[3] = {
+        0x5e5f050f,  // syscall; pop %arg1{%rdi}; pop %arg2{%rsi}
+        0xff3e585a,  // pop %arg3{%rdx}; pop %rax; notrack jmp
+        0x909090e0   // *%rax; nop; nop; nop
+    };
     DPRINTF("make_hatch %%p %%p %%x\\n", phdr, next_unc, frag_mask);
     if (phdr->p_type==PT_LOAD && phdr->p_flags & PF_X) {
         next_unc += phdr->p_memsz - phdr->p_filesz;  // Skip over local .bss
         frag_mask &= -(long)next_unc;  // bytes left on page
-        if (6 <= frag_mask) {
-            hatch = next_unc;
-            (( long *)hatch)[0] = 0x5e5f050f;  // syscall; pop %arg1{%rdi}; pop %arg2{%rsi}
-            ((short *)hatch)[2] = 0xc35a;  // pop %arg3{%rdx}; ret
+        if (sizeof(code) <= frag_mask) {
+            hatch = (unsigned *)next_unc;
+            hatch[0] = code[0];
+            hatch[1] = code[1];
+            hatch[2] = code[2];
         }
         else { // Does not fit at hi end of .text, so must use a new page "permanently"
-            int mfd = memfd_create(addr_string("upx"), 0);  // the directory entry
-            Pwrite(mfd, addr_string("\x0f\x05\x5f\x5e\x5a\xc3"), 6);
-            hatch = Pmap(0, 6, PROT_READ|PROT_EXEC, MAP_PRIVATE, mfd, 0);
+            unsigned long fdmap = (long)upx_mmap_and_fd((void *)0, sizeof(code), nullptr);
+            unsigned mfd = -1+ (0xfff& fdmap);
+            write(mfd, &code, sizeof(code));
+            hatch = mmap((void *)(fdmap & ~0xffful), sizeof(code),
+              PROT_READ|PROT_EXEC, MAP_PRIVATE, mfd, 0);
             close(mfd);
         }
     }
     DPRINTF("hatch=%%p\\n", hatch);
-    return hatch;
+    return (char *)hatch;
 }
 #elif defined(__powerpc64__)  //}{
 static unsigned
@@ -278,8 +316,8 @@ ORRX(unsigned ra, unsigned rs, unsigned rb) // or ra,rs,rb
 }
 
 static char *
-make_hatch_ppc64(
-    Elf64_Phdr const *const phdr,
+make_hatch(
+    ElfW(Phdr) const *const phdr,
     char *next_unc,
     unsigned frag_mask
 )
@@ -312,10 +350,47 @@ make_hatch_ppc64(
     DPRINTF("hatch=%%p\\n", hatch);
     return hatch;
 }
-#elif defined(__aarch64__)  //{
+#elif defined(__riscv)  //}{
+extern void *my_memcpy(void *, const void *, long unsigned int);
+
+static void * __attribute__((noinline))
+make_hatch(
+    ElfW(Phdr) const *const phdr,
+    char *next_unc,
+    unsigned const frag_mask
+)
+{
+    unsigned long a = (unsigned long)next_unc;
+    short *hatch = (short *)((1u & a) + a);
+#define SZ_CODE (8*sizeof(short))
+    char const *code;  // embedded "\x00" terminates!
+    asm("jal %0,0f; \
+        .int 0x00000073; \
+        .short 0x6502, 0x65a2, 0x6642, 0x60e2, 0x6105; \
+        .short 0x8082; \
+     0: .balign 4" \
+/*out*/ : "=r"(code) \
+/* in*/ : \
+/*und*/ : );
+    DPRINTF("make_hatch %%p %%p %%x\\n", phdr, next_unc, frag_mask);
+    if (phdr->p_type==PT_LOAD && phdr->p_flags & PF_X) {
+        if (SZ_CODE <= (unsigned)(frag_mask & -(long)hatch)) {
+            my_memcpy(hatch, code, SZ_CODE);
+        }
+        else { // Does not fit at hi end of .text, so must use a new page "permanently"
+            int mfd = memfd_create(addr_string("upx"), MFD_EXEC);  // the directory entry
+            write(mfd, &code, SZ_CODE);
+            hatch = mmap(0, SZ_CODE, PROT_READ|PROT_EXEC, MAP_SHARED, mfd, 0);
+            close(mfd);
+        }
+    }
+    DPRINTF("hatch=%%p\\n", hatch);
+    return hatch;
+}
+#elif defined(__aarch64__)  //}{
 static char *
-make_hatch_arm64(
-    Elf64_Phdr const *const phdr,
+make_hatch(
+    ElfW(Phdr) const *const phdr,
     char *next_unc,
     unsigned frag_mask
 )
@@ -365,57 +440,124 @@ make_hatch_arm64(
          |(REP8(PROT_WRITE) & EXP8(PF_W)) \
         ) >> ((pf & (PF_R|PF_W|PF_X))<<2) ))
 
-#define nullptr (void *)0
-
 #undef PAGE_MASK
-static unsigned long
-get_PAGE_MASK(void)  // the mask which KEEPS the page, discards the offset
+
+#if defined(__riscv)  //{  why is riscv the only one?
+extern ElfW(Addr) get_page_mask(void);
+#else  //}{
+static ElfW(Addr)
+get_page_mask(void)  // the mask which KEEPS the page, discards the offset
 {
-    unsigned long rv = ~0xffful;  // default to (PAGE_SIZE == 4KiB)
+    ElfW(Addr) rv = ~0xffful;  // default to (PAGE_SIZE == 4KiB)
     int fd = openat(0, addr_string("/proc/self/auxv"), O_RDONLY, 0);
     if (0 <= fd) {
-        Elf64_auxv_t data[40];
-        Elf64_auxv_t *end = &data[read(fd, data, sizeof(data)) / sizeof(data[0])];
+        ElfW(auxv_t) data[40];
+        ElfW(auxv_t) *end = &data[read(fd, data, sizeof(data)) / sizeof(data[0])];
         close(fd);
-        Elf64_auxv_t *ptr; for (ptr = &data[0]; ptr < end ; ++ptr) {
+        ElfW(auxv_t) *ptr; for (ptr = &data[0]; ptr < end ; ++ptr) {
             if (AT_PAGESZ == ptr->a_type) {
                 rv = (0u - ptr->a_un.a_val);
                 break;
             }
         }
     }
-    DPRINTF("get_PAGE_MASK= %%p\\n", rv);
+    DPRINTF("get_page_mask= %%p\\n", rv);
     return rv;
 }
+#endif  //}
 
 extern void *memcpy(void *dst, void const *src, size_t n);
-extern void *memset(void *dst, unsigned val, size_t n);
+extern void *memset(void *dst, int val, size_t n);
+
+// maximum page sizes
+#if defined(__powerpc64__) || defined(__powerpc__)
+#define SAVED_SIZE (1<<16)  /* 64 KB */
+#else
+#define SAVED_SIZE (1<<14)  /* 16 KB */
+#endif
 
 #ifndef __arm__  //{
 // Segregate large local array, to avoid code bloat due to large displacements.
 static void
-underlay(unsigned size, char *ptr, unsigned len)  // len <= PAGE_SIZE
+underlay(unsigned size, char *ptr, unsigned page_mask)
 {
-    DPRINTF("underlay size=%%u  ptr=%%p  len=%%u\\n", size, ptr, len);
-    unsigned saved[4096/sizeof(unsigned)];
-    memcpy(saved, ptr, len);
-    mmap(ptr, size, PROT_WRITE|PROT_READ,
-        MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
-    memcpy(ptr, saved, len);
+    unsigned frag = ~page_mask & (unsigned)(long)ptr;
+    if (frag) {
+        unsigned saved[SAVED_SIZE / sizeof(unsigned)];  // want alignment
+        ptr -= frag;
+        memcpy(saved, ptr, frag);
+        mmap(ptr, frag + size, PROT_WRITE|PROT_READ,
+            MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+        memcpy(ptr, saved, frag);
+    }
+    else { // already page-aligned
+        mmap(ptr, frag + size, PROT_WRITE|PROT_READ,
+            MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+    }
 }
-#else  //}{
+#else  //}{ // use assembler because large local array on __arm__ is horrible
 extern void
-underlay(unsigned size, char *ptr, unsigned len);
+underlay(unsigned size, char *ptr, unsigned page_mask);
 #endif  //}
 
 // Exchange the bits with values 4 (PF_R, PROT_EXEC) and 1 (PF_X, PROT_READ)
 // Use table lookup into a PIC-string that pre-computes the result.
-unsigned PF_to_PROT(Elf64_Phdr const *phdr)
+unsigned PF_to_PROT(ElfW(Phdr) const *phdr)
 {
     return 7& addr_string("@\x04\x02\x06\x01\x05\x03\x07")
         [phdr->p_flags & (PF_R|PF_W|PF_X)];
 }
 
+unsigned
+fini_SELinux(
+    unsigned size,
+    char *ptr,
+    ElfW(Phdr) const *phdr,
+    unsigned mfd,
+    ElfW(Addr) base
+)
+{
+    DPRINTF("fini_SELinux  size=%%p  ptr=%%p  phdr=%%p  mfd=%%p  base=%%p\\n",
+            size, ptr, phdr, mfd, base);
+    if (phdr->p_flags & PF_X) {
+        // Map the contents of mfd as per *phdr.
+
+        Psync(ptr, size, MS_SYNC); // be sure file gets de-compressed bytes
+            // Android 14 gets -EINVAL; ignore it
+
+        Punmap(ptr, size);
+        Pmap(ptr, size, PF_to_PROT(phdr), MAP_FIXED|MAP_PRIVATE, mfd, 0);
+        close(mfd);
+    }
+    else { // easy
+        Pprotect( (char *)(phdr->p_vaddr + base), phdr->p_memsz, PF_to_PROT(phdr));
+    }
+    return 0;
+}
+
+unsigned
+prep_SELinux(unsigned size, char *ptr, ElfW(Addr) page_mask) // returns mfd
+{
+    // Cannot set PROT_EXEC except via mmap() into a region (Linux "vma")
+    // that has never had PROT_WRITE.  So use a Linux-only "memory file"
+    // to hold the contents.
+    unsigned saved[SAVED_SIZE / sizeof(unsigned)];  // want alignment
+    char *page = (char *)(page_mask & (ElfW(Addr))ptr);
+    unsigned frag = (unsigned)(ptr - page);
+    if (frag) {
+        memcpy(saved, page, frag);
+    }
+    char *val = upx_mmap_and_fd(page, frag + size, nullptr);
+    unsigned mfd = 0xfff & (unsigned)(ElfW(Addr))val;
+    val -= mfd; --mfd;
+    if (val != page) {
+        my_bkpt((void const *)0x1262, val, page, ptr, frag);
+    }
+    if (frag)
+        write(mfd, saved, frag);  // Save lo fragment of contents on page.
+    return mfd;
+}
+
 typedef struct {
     long argc;
     char **argv;
@@ -423,7 +565,7 @@ typedef struct {
 } So_args;
 
 typedef struct {
-    unsigned off_reloc;  // distance back to &Elf64_Ehdr
+    unsigned off_reloc;  // distance back to &ElfW(Ehdr)
     unsigned off_user_DT_INIT;
     unsigned off_xct_off;  // where un-compressed bytes end
     unsigned off_info;  //  xct_off: {l_info; p_info; b_info; compressed data)
@@ -437,10 +579,10 @@ void *
 upx_so_main(  // returns &escape_hatch
     So_info *so_info,
     So_args *so_args,
-    Elf64_Ehdr *elf_tmp  // scratch for Elf64_Ehdr and Elf64_Phdrs
+    ElfW(Ehdr) *elf_tmp  // scratch for ElfW(Ehdr) and ElfW(Phdrs)
 )
 {
-    unsigned long const PAGE_MASK = get_PAGE_MASK();
+    ElfW(Addr) const page_mask = get_page_mask();
     char *const va_load = (char *)&so_info->off_reloc - so_info->off_reloc;
     So_info so_infc;  // So_info Copy
     memcpy(&so_infc, so_info, sizeof(so_infc));  // before de-compression overwrites
@@ -457,7 +599,7 @@ upx_so_main(  // returns &escape_hatch
     // Copy compressed data before de-compression overwrites it.
     char *const sideaddr = mmap(nullptr, cpr_len, PROT_WRITE|PROT_READ,
         MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
-    DPRINTF("sideaddr=%%p\\n", sideaddr);
+    DPRINTF("&sideaddr=%%p\\n", &sideaddr);
     memcpy(sideaddr, cpr_ptr, cpr_len);
 
     // Transition to copied data
@@ -479,110 +621,66 @@ upx_so_main(  // returns &escape_hatch
     //   De-compress from remaining [sideaddr, +sidelen).
     //   Pprotect(,, PF_TO_PROT(.p_flags));
 
-    // Get the uncompressed Elf64_Ehdr and Elf64_Phdr
+    // Get the uncompressed ElfW(Ehdr) and ElfW(Phdr)
     // The first b_info is aligned, so direct access to fields is OK.
     Extent x1 = {binfo->sz_unc, (char *)elf_tmp};  // destination
     Extent x0 = {binfo->sz_cpr + sizeof(*binfo), (char *)binfo};  // source
     unpackExtent(&x0, &x1);  // de-compress _Ehdr and _Phdrs; x0.buf is updated
 
-    Elf64_Phdr const *phdr = (Elf64_Phdr *)(1+ elf_tmp);
-    Elf64_Phdr const *const phdrN = &phdr[elf_tmp->e_phnum];
-    while (phdr->p_type != PT_LOAD) ++phdr;  // skip PT_PHDR if any
-    Elf64_Addr const base = (Elf64_Addr)va_load - phdr->p_vaddr;
-    DPRINTF("base=%%p\\n", base);
-
-    if (phdr->p_flags & PF_X) {
-        int mfd = memfd_create(addr_string("upx"), 0);
-        unsigned mfd_len = 0ul - PAGE_MASK;
-        Pwrite(mfd, elf_tmp, binfo->sz_unc);  // de-compressed Elf_Ehdr and Elf_Phdrs
-        Pwrite(mfd, binfo->sz_unc + va_load, mfd_len - binfo->sz_unc);  // rest of 1st page
-
-        Punmap(va_load, mfd_len);  // make SELinux forget any previous protection
-        Elf64_Addr va_mfd = (Elf64_Addr)Pmap(va_load, mfd_len, PF_to_PROT(phdr),
-            MAP_FIXED|MAP_PRIVATE, mfd, 0); (void)va_mfd;
-        close(mfd);
-    }
+    ElfW(Phdr) const *phdr = (ElfW(Phdr) *)(1+ elf_tmp);
+    ElfW(Phdr) const *const phdrN = &phdr[elf_tmp->e_phnum];
 
     // Process each read-only PT_LOAD.
     // A read+write PT_LOAD might be relocated by rtld before de-compression,
     // so it cannot be compressed.
-    struct b_info al_bi;  // for aligned data from binfo
     void *hatch = nullptr;
+    ElfW(Addr) base = 0;
+    int n_load = 0;
 
     for (; phdr < phdrN; ++phdr)
-    if ( phdr->p_type == PT_LOAD && !(phdr->p_flags & PF_W)) {
-        DPRINTF("phdr@%%p  p_offset=%%p  p_vaddr=%%p  p_filesz=%%p  p_memsz=%%p  binfo=%%p\\n",
-            phdr, phdr->p_offset, phdr->p_vaddr, phdr->p_filesz, phdr->p_memsz, x0.buf);
-
-        if ((phdr->p_filesz + phdr->p_offset) <= so_infc.off_xct_off) {
-            continue;  // below compressed region
+    if (phdr->p_type == PT_LOAD && !(phdr->p_flags & PF_W)) {
+        if  (!base) {
+            base = (ElfW(Addr))va_load - phdr->p_vaddr;
+            DPRINTF("base=%%p\\n", base);
         }
-        Elf64_Addr const pfx = (so_infc.off_xct_off < phdr->p_offset)
-            ? 0  // entire PT_LOAD is compressed
-            : so_infc.off_xct_off - phdr->p_offset ;  // below xct_off is not
+        unsigned const va_top = phdr->p_filesz + phdr->p_vaddr;
+        // Need un-aligned read of b_info to determine compression sizes.
+        struct b_info al_bi;  // for aligned data from binfo
         x0.size = sizeof(struct b_info);
         xread(&x0, (char *)&al_bi, x0.size);  // aligned binfo
         x0.buf -= sizeof(al_bi);  // back up (the xread() was a peek)
-        DPRINTF("next1 pfx=%%x binfo@%%p (%%p %%p %%p)\\n", pfx, x0.buf,
-            al_bi.sz_unc, al_bi.sz_cpr, *(unsigned *)(void *)&al_bi.b_method);
+        x0.size = al_bi.sz_cpr;
+        x1.size = al_bi.sz_unc;
+        x1.buf = (void *)(va_top + base - al_bi.sz_unc);
 
-        // Using .p_memsz implicitly handles .bss via MAP_ANONYMOUS.
-        // Omit any non-tcompressed prefix (below xct_off)
-        x1.buf =  (char *)(pfx + phdr->p_vaddr + base);
-        x1.size = phdr->p_memsz - pfx;
+        DPRINTF("\\nphdr@%%p  p_offset=%%p  p_vaddr=%%p  p_filesz=%%p  p_memsz=%%p\\n",
+            phdr, phdr->p_offset, phdr->p_vaddr, phdr->p_filesz, phdr->p_memsz);
+        DPRINTF("x0=%%p  x1=%%p\\n", &x0, &x1);
 
-        unsigned const frag = (phdr->p_vaddr + pfx) & ~PAGE_MASK;  // lo fragment on page
-        x1.buf  -= frag;
-        x1.size += frag;
-        DPRINTF("phdr(%%p %%p) xct_off=%%x frag=%%x\\n", x1.buf, x1.size, xct_off, frag);
+        if ((phdr->p_filesz + phdr->p_offset) <= xct_off) { // va_top <= xct_off
+            if (!n_load) {
+                ++n_load;
+                continue;  // 1st PT_LOAD is non-compressed loader tables ONLY!
+            }
+        }
 
         int mfd = 0;
-        char *mfd_addr = 0;
-        if (phdr->p_flags & PF_X) { // SELinux
-            // Cannot set PROT_EXEC except via mmap() into a region (Linux "vma")
-            // that has never had PROT_WRITE.  So use a Linux-only "memory file"
-            // to hold the contents.
-            mfd = memfd_create(addr_string("upx"), 0);  // the directory entry
-            ftruncate(mfd, x1.size);  // Allocate the pages in the file.
-            Pwrite(mfd, x1.buf, frag);  // Save lo fragment of contents on first page.
-            Punmap(x1.buf, x1.size);
-            mfd_addr = Pmap(x1.buf, x1.size, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_SHARED, mfd, 0);
-            DPRINTF("mfd_addr= %%p\\n", mfd_addr);  // Re-use the address space
+        if (phdr->p_flags & PF_X) {
+            mfd = prep_SELinux(x1.size, x1.buf, page_mask);
         }
         else {
-            underlay(x1.size, x1.buf, frag);  // also makes PROT_WRITE
+            underlay(x1.size, x1.buf, page_mask);  // also makes PROT_WRITE
         }
-
-        x1.buf += frag;
-        x1.size = al_bi.sz_unc;
-        x0.size = al_bi.sz_cpr + sizeof(struct b_info);
-        DPRINTF("before unpack x0=(%%p %%p  x1=(%%p %%p)\\n", x0.size, x0.buf, x1.size, x1.buf);
-        unpackExtent(&x0, &x1);  // updates x0 and x1
-        DPRINTF(" after unpack x0=(%%p %%p  x1=(%%p %%p)\\n", x0.size, x0.buf, x1.size, x1.buf);
-
+        Extent xt = x1;
+        unpackExtent(&x0, &x1);
         if (!hatch && phdr->p_flags & PF_X) {
-#if defined(__x86_64)  //{
-            hatch = make_hatch_x86_64(phdr, x1.buf, ~PAGE_MASK);
-#elif defined(__powerpc64__)  //}{
-            hatch = make_hatch_ppc64(phdr, x1.buf, ~PAGE_MASK);
-#elif defined(__aarch64__)  //}{
-            hatch = make_hatch_arm64(phdr, x1.buf, ~PAGE_MASK);
-#endif  //}
-        }
-
-        if (phdr->p_flags & PF_X) { // SELinux
-            // Map the contents of mfd as per *phdr.
-            DPRINTF("mfd mmap addr=%%p  len=%%p\\n", (phdr->p_vaddr + base + pfx), al_bi.sz_unc);
-            Punmap(mfd_addr, frag + al_bi.sz_unc);  // Discard RW mapping; mfd has the bytes
-            Pmap((char *)(phdr->p_vaddr + base + pfx), al_bi.sz_unc, PF_to_PROT(phdr),
-                MAP_FIXED|MAP_PRIVATE, mfd, 0);
-            close(mfd);
-        }
-        else { // easy
-            Pprotect( (char *)(phdr->p_vaddr + base), phdr->p_memsz, PF_to_PROT(phdr));
+            hatch = make_hatch(phdr, x1.buf, ~page_mask);
+            fini_SELinux(xt.size, xt.buf, phdr, mfd, base);
         }
+        ++n_load;
     }
 
+    DPRINTF("Punmap sideaddr=%%p  cpr_len=%%p\\n", sideaddr, cpr_len);
     Punmap(sideaddr, cpr_len);
     DPRINTF("calling user DT_INIT %%p\\n", dt_init);
     dt_init(so_args->argc, so_args->argv, so_args->envp);
